--[[
Name: LibSQL-1.0
Revision: $Rev: 6 $
Author(s): Humbedooh
Description: SQL library for Lua
Optional dependencies: LibStub
License: All Sporks Reserved!
]]
-----------------------------------------------------------------------

local LIBSQL = "LibSQL-1.0"
local LIBSQL_MINOR = tonumber(("$Rev: 06 $"):match("(%d+)")) or 10000
local LibSQL = false
if LibStub then
    LibSQL = LibStub:NewLibrary(LIBSQL, LIBSQL_MINOR)
    _G.LibSQL = LibSQL
else
    if not _G.LibSQL or (type(_G.LibSQL) == "table" and _G.LibSQL.rev < LIBSQL_MINOR) then
        _G.LibSQL = {rev = LIBSQL_MINOR}
    end
    LibSQL = _G.LibSQL
end
if not LibSQL then return end
LibSQL_Functions = {'ABS', 'ACOS', 'ADDDATE', 'ASCII', 'ASIN', 'ATAN', 'ATAN2', 'AVG', 'BIN', 'BIT_AND', 'BIT_COUNT', 'BIT_LENGTH', 'BIT_OR', 'CAST', 'CEIL', 'CEILING', 'CHAR_LENGTH', 'CHARACTER_LENGTH', 'CONCAT', 'CONCAT_WS', 'CONV', 'CONVERT', 'COS', 'COT', 'COUNT', 'CURDATE', 'CURRENT_DATE', 'CURRENT_TIME', 'CURRENT_TIMESTAMP', 'CURTIME', 'DATE_ADD', 'DATE_FORMAT', 'DATE_SUB', 'DAYNAME', 'DAYOFMONTH', 'DAYOFWEEK', 'DAYOFYEAR', 'DECODE', 'DEGREES', 'ENCODE', 'ENCRYPT', 'EXP', 'EXPORT_SET', 'EXTRACT', 'FIELD', 'FIND_IN_SET', 'FLOOR', 'FORMAT', 'FOUND_ROWS', 'FROM_DAYS', 'FROM_UNIXTIME', 'GREATEST', 'HEX', 'IFNULL', 'INTERVAL', 'ISNULL', 'LAST_INSERT_ID', 'LCASE', 'LEAST', 'LEFT', 'LENGTH', 'LN', 'LOCATE', 'LOG', 'LOG2', 'LOG10', 'LOWER', 'LPAD', 'LTRIM', 'MAKE_SET', 'MAX', 'MD5', 'MID', 'MIN', 'MOD', 'MONTHNAME', 'NOW', 'NULLIF', 'OCT', 'OCTET_LENGTH', 'ORD', 'PASSWORD', 'PERIOD_ADD', 'PERIOD_DIFF', 'PI', 'POSITION', 'POW', 'POWER', 'QUARTER', 'QUOTE', 'RADIANS', 'RAND', 'RELEASE_LOCK', 'REPEAT', 'REVERSE', 'RIGHT', 'ROUND', 'RPAD', 'RTRIM', 'SEC_TO_TIME', 'SHA', 'SHA1', 'SIGN', 'SIN', 'SPACE', 'SQRT', 'STD', 'STDDEV', 'STRCMP', 'SUBDATE', 'SUBSTRING', 'SUBSTRING_INDEX', 'SUM', 'SYSDATE', 'TAN', 'TIME_FORMAT', 'TIME_TO_SEC', 'TO_DAYS', 'TRIM', 'UCASE', 'UNIX_TIMESTAMP', 'UPPER', 'WEEK', 'WEEKDAY', 'YEARWEEK'}
LibSQL_Field_Flags = {'AUTO_INCREMENT','BINARY','DEFAULT','NATIONAL','PRECISION','UNSIGNED','VARYING','ZEROFILL', 'NOTNIL'}
LibSQL_Reserved = {'ACTION','ADD','AFTER','AGAINST','AGGREGATE','ALL','ALTER','ANALYSE','ANALYZE','AND','AS','ASC','AUTOCOMMIT','AVG_ROW_LENGTH','BACKUP','BEGIN','BETWEEN','BOTH','BY','CASCADE','CASE','CHANGE','CHANGED','CHECK','CHECKSUM','CLIENT','COLUMN','COLUMNS','COMMENT','COMMIT','COMMITTED','COMPRESSED','CONCURRENT','CONSTRAINT','CREATE','CROSS','DATA','DATABASE','DATABASES','DAY','DAY_HOUR','DAY_MINUTE','DAY_SECOND','DELAYED','DELAY_KEY_WRITE','DELETE','DESC','DESCRIBE','DISTINCT','DISTINCTROW','DO','DROP','DUMPFILE','DYNAMIC','ELSE','ENCLOSED','END','ESCAPE','ESCAPED','EXECUTE','EXISTS','EXPLAIN','EXTENDED','FAST','FIELDS','FILE','FIRST','FIXED','FLUSH','FOR','FOREIGN','FROM','FULL','FULLTEXT','FUNCTION','GEMINI','GEMINI_SPIN_RETRIES','GLOBAL','GRANT','GRANTS','GROUP','HAVING','HEAP','HIGH_PRIORITY','HOSTS','HOUR','HOUR_MINUTE','HOUR_SECOND','IDENTIFIED','IF','IGNORE','IN','INDEX','INDEXES','INFILE','INNER','INSERT','INSERT_ID','INSERT_METHOD','INTERVAL','INTO','IS','ISOLATION','JOIN','KEY','KEYS','KILL','LAST_INSERT_ID','LEADING','LEFT','LEVEL','LIKE','LIMIT','LINES','LOAD','LOCAL','LOCK','LOCKS','LOGS','LOW_PRIORITY','MASTER','MASTER_CONNECT_RETRY','MASTER_HOST','MASTER_LOG_FILE','MASTER_LOG_POS','MASTER_PASSWORD','MASTER_PORT','MASTER_USER','MATCH','MAX_CONNECTIONS_PER_HOUR','MAX_QUERIES_PER_HOUR','MAX_ROWS','MAX_UPDATES_PER_HOUR','MEDIUM','MERGE','MIN_ROWS','MINUTE','MINUTE_SECOND','MODE','MODIFY','MONTH','MRG_MYISAM','MYISAM','NATURAL','NO','NOT','NULL','ON','OPEN','OPTIMIZE','OPTION','OPTIONALLY','OR','ORDER','OUTER','OUTFILE','PACK_KEYS','PARTIAL','PASSWORD','PRIMARY','PRIVILEGES','PROCEDURE','PROCESS','PROCESSLIST','PURGE','QUICK','RAID0','RAID_CHUNKS','RAID_CHUNKSIZE','RAID_TYPE','READ','REFERENCES','REGEXP','RELOAD','RENAME','REPAIR','REPEATABLE','REPLACE','REPLICATION','RESET','RESTORE','RESTRICT','RETURNS','REVOKE','RIGHT','RLIKE','ROLLBACK','ROW','ROW_FORMAT','ROWS','SECOND','SELECT','SERIALIZABLE','SESSION','SHARE','SHOW','SHUTDOWN','SLAVE','SONAME','SQL_AUTO_IS_NULL','SQL_BIG_RESULT','SQL_BIG_SELECTS','SQL_BIG_TABLES','SQL_BUFFER_RESULT','SQL_LOG_BIN','SQL_LOG_OFF','SQL_LOG_UPDATE','SQL_LOW_PRIORITY_UPDATES','SQL_MAX_JOIN_SIZE','SQL_QUOTE_SHOW_CREATE','SQL_SAFE_UPDATES','SQL_SELECT_LIMIT','SQL_SLAVE_SKIP_COUNTER','SQL_SMALL_RESULT','SQL_WARNINGS','START','STARTING','STATUS','STOP','STRAIGHT_JOIN','STRING','STRIPED','SUPER','TABLE','TABLES','TEMPORARY','TERMINATED','THEN','TO','TRAILING','TRANSACTION','TRUNCATE','TYPE','UNCOMMITTED','UNION','UNIQUE','UNLOCK','UPDATE','USAGE','USE','USING','VALUES','VARIABLES','WHEN','WHERE','WITH','WORK','WRITE','YEAR_MONTH'}
LibSQL_Field_Types = {'INTEGER', 'NUMBER', 'STRING', 'BOOLEAN', 'ENUM', 'DATE', 'TIMESTAMP', 'BLOB'}

-- [[ UPVALUES ]]--
local _G = _G
local assert = _G.assert
local time = assert(_G.time or _G.os.time)
local upper = assert(_G.string.upper)
local lower = assert(_G.string.lower)
local match = assert(_G.string.match)
local gmatch = assert(_G.string.gmatch)
local gsub = assert(_G.string.gsub)
local sub = assert(_G.string.sub)
local find = assert(_G.string.find)
local format = assert(_G.string.format)
local tonumber = assert(_G.tonumber)
local tostring = assert(_G.tostring)
local format = assert(_G.string.format)
local getmetatable = assert(_G.getmetatable)
local setmetatable = assert(_G.setmetatable)
local pairs = assert(_G.pairs)
local next = assert(_G.next)
local type = assert(_G.type)
local rawget = assert(_G.rawget)
local error = assert(_G.error)
local select = assert(_G.select)
local tsort = assert(_G.table.sort)
local tinsert = assert(_G.table.insert)
local tremove = assert(_G.table.remove)
local floor = assert(_G.math.floor)

--[[ TOOL SET ]]--
local function LibSQL_Validate_Field_Name(name) if type(name) ~= "string" or match(name, "[^a-zA-Z0-9_]") then return false else return true end end
local function LibSQL_Set_Weak_Table(tbl) setmetatable(tbl, {__mode = "v"}) end
local function LibSQL_Field_Exists(tbl, field) local t,x = tbl._fields, lower(field) for n = 1, #t do if t[n].id == x then return t[n] end end return nil end
local function LibSQL_Field_Valid_Type(field) local field = upper(field) for n = 1, #LibSQL_Field_Types do if LibSQL_Field_Types[n] == field then return true end end return false end
local function LibSQL_Field_Valid_Flag(field) local field = upper(field) for n = 1, #LibSQL_Field_Flags do if LibSQL_Field_Flags[n] == field then return true end end return false end
local function LibSQL_Parse_Alter(statement)
    local ret = {}
    local cmd = match(statement, "^%s*([%w]+)")
    if cmd then
        ret.cmd = upper(cmd)
        local a,b = find(statement, "(%b())")
        if a and b then statement = sub(statement, a+1,b-1)
        else statement = gsub(statement, "^%s*[%w]+%s*", "", 1)
        end
        for el in gmatch(statement, "([^,]+)%s*,?%s*") do
            ret[#ret+1] = {match(el, "(%w+)%s*(%w*)%s*(.*)")}
        end
    end
    return ret
end
LibSQL_Parse_Alter("Add (something string, someother)")


--[[ RESOURCE OBJECTs ]]--
LibSQL_Resource = {}

local function LibSQL_Resource_Next(res)
    res._pointer = (res._pointer or 0)+1
    local entry = res._entries[res._pointer]
    if not entry then return nil end
    return unpack(entry or {})
end

local function LibSQL_Resource_Next_Arrayed(res)
    res._pointer = (res._pointer or 0)+1
    local entry = res._entries[res._pointer]
    if not entry then return nil end
    return entry or {}
end

local function LibSQL_Resource_Next_Objectified(res)
    res._pointer = (res._pointer or 0)+1
    local entry = res._entries[res._pointer]
    if not entry then return nil end
    local obj = {}
    local f = res._fields
    for n = 1, #f do
        obj[f[n].key] = entry[n]
    end
    return obj
end

function LibSQL_Resource:Flush() self._pointer = nil self._entries = {} self._fields = {} end
function LibSQL_Resource:Rewind() self._pointer = nil end
function LibSQL_Resource:Seek(num) self._pointer = tonumber(num) return self._pointer end
function LibSQL_Resource:Rows() return #self._entries end
function LibSQL_Resource:Fetch() return LibSQL_Resource_Next, self, nil end
function LibSQL_Resource:FetchObject() return LibSQL_Resource_Next_Objectified, self, nil end
function LibSQL_Resource:FetchArray() return LibSQL_Resource_Next_Arrayed, self, nil end

function LibSQL_Resource:New(hash)
    local res = {}
    if type(hash) ~= "table" then hash = {} end
    res._fields =   hash._fields or LibSQL_Set_Weak_Table({})
    res._entries =  hash._entries or LibSQL_Set_Weak_Table({})
    res._pointer = nil
    res.Fetch =             LibSQL_Resource.Fetch
    res.FetchObject =       LibSQL_Resource.FetchObject
    res.FetchArray =        LibSQL_Resource.FetchArray
    res.Rewind =            LibSQL_Resource.Rewind
    res.Seek =              LibSQL_Resource.Seek
    res.Rows =              LibSQL_Resource.Rows
    res.Flush =				LibSQL_Resource.Flush
    return res
end


--[[ DATABASE OBJECTS ]]--

-- example:
-- CREATE TABLE myTable (name STRING UNIQUE, level NUMBER, id NUMBER AUTO_INCREMENT);
-- Translates to;
-- db:Create("myTable"):Alter("ADD (name STRING UNIQUE, level NUMBER, id NUMBER AUTO_INCREMENT)")
function LibSQL:_Create(tbl)
    if LibSQL_Validate_Field_Name(tbl) then
        local db = self
        local tblint = lower(tbl)
        if db._tables[tblint] then error(format(tbl,"LibSQL: Table '%s' already exists!")) return end
        db._tables[tblint] = {created = time(), name = tbl, insert_id = 0, _fields = {}, _entries = {}}
        return db._tables[tblint]
    end
end


function LibSQL:_Alter(statement)
    local args = LibSQL_Parse_Alter(statement)
    local tbl = self
    local cmd = args.cmd;
    if 		cmd == "ADD"    then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' already exists!", args[n][1])) return
            else
                local t,f = upper(args[n][2]), upper(args[n][3])
                if not LibSQL_Field_Valid_Type(t) then error(format("LibSQL: Invalid field type '%s'!", t)) return end
                if not LibSQL_Field_Valid_Flags(f) then error(format("LibSQL: Invalid field flag '%s'!", f)) return end
            end
        end
        for n = 1, #args do
            local n,t,f = args[n][1], upper(args[n][2]), upper(args[n][3])
            tinsert(tbl._fields, {id=lower(n), name=n, flags=f})
        end
    elseif	cmd == "DROP"   then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if not LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' does not exist!", args[n][1])) return end
        end
        local t = tbl._fields
        for n = 1, #args do
            local x = lower(args[n][2]) for n = 1, #t do if t[n].id == x then tremove(t, n) break end end
        end
    elseif cmd == "RENAME"	then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if not LibSQL_Validate_Field_Name(args[n][2]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][2])) return end
            if not LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' does not exist!", args[n][1])) return end
            if not LibSQL_Field_Exists(tbl, args[n][2]) then error(format("LibSQL: Field '%s' already exist!", args[n][2])) return end
        end
        for n = 1, #args do
            local x = lower(args[n][2]) for n = 1, #t do if t[n].id == x then t[n].id = x t[n].name = args[n][2] break end end
        end
    elseif cmd == "MODIFY"	then
        for n = 1, #args do
            if not LibSQL_Validate_Field_Name(args[n][1]) then error(format("LibSQL: Field name '%s' is invalid!", args[n][1])) return end
            if not LibSQL_Validate_Field_Type(args[n][2]) then error(format("LibSQL: Field type '%s' is invalid!", args[n][2])) return end
            if not LibSQL_Validate_Field_Flags(args[n][3]) then error(format("LibSQL: Field flag '%s' is invalid!", args[n][3])) return end
            if not LibSQL_Field_Exists(tbl, args[n][1]) then error(format("LibSQL: Field '%s' does not exist!", args[n][1])) return end
        end
        for n = 1, #args do
            local x = lower(args[n][1]) for n = 1, #t do if t[n].id == x then t[n].type = upper(args[n][2]) break end end
        end
    end
end

function LibSQL:_Get(tbl)
    local db = self
    if LibSQL_Validate_Field_Name(tbl) then
        local tblhash = db._tables[lower(tbl)]
        if tblhash then return LibSQL_Resource:New(tblhash) end
    end
    return LibSQL_Resource:New()
end

function LibSQL:Parse(arg)
    local db = self
    if type(arg) ~= "string" or type(db) ~= "table" then return end
    local cmd = arg:match("^%s*(%a+)%s")
    if cmd then
        cmd = cmd:upper()
    end
end


function LibSQL:Open(db)
    if type(db) ~= "table" or not db._tables then error("Usage: LibSQL:Open(database)") return end
    db.Execute = 	LibSQL.Execute
    db._Get = 		LibSQL._Get
    db._Alter =		LibSQL._Alter
    db._Create =	LibSQL._Create
    return db
end

function LibSQL:New() return LibSQL:Open({_tables = {}}) end

--db:_Get("myTable"):_Find("name=Moo"):_Sort("name ASC"):_Limit(0,10):_Select("name, level")
--[[

local db = LibSQL_Resource:New()
print("got", db:Rows(), "results!")
for a,b,c,d in db:FetchObject() do
    print(a.name, a.level)
end

]]--

